/*******************************************************************************
* 
*	A test MPID implementation for Windows NT:
* 
* WINNT_init.c: 
* contains tentative versions of 
*
* int MPID_WINNT_Init (int*, char***)
*
* int MPID_WINNT_WorldSize(int*)
*
* int MPID_WINNT_MyRank(int*)
*
* void MPID_WINNT_End(void)
*
* functions. The code is self-explanatory. Some comments clarify details.
* The code is based on MPID T3D and Cameleon implementations (which means that 
* portions of T3D and Cameleon code were used).
*
* Programmer: Boris V. Protopopov. 
* Mississippi State University, 
* ICDCR Lab.
* 
* Last update: 
* 02.06.96.
*
********************************************************************************/                

#include <stdio.h>
#include <windows.h>
#include <windowsx.h>
#include <crtdbg.h>
#include "packets.h"
#include "queue.h"
/* Global variables which will be used by send/recv functions */

int MPID_WorldSize;
int MPID_MyWorldRank;
int MPID_NumShmemProcs;

MPID_WINNT_SHMEM_BUF_T** WINNT_ShmemBufs;

HANDLE hEndProcessEvent;

HANDLE hTerminatorEvent;

HANDLE hThreads[3];

static HANDLE hBarrierEvent;
static HANDLE hBarrierSemaphore;

/* Function definitions */

void* MPID_WINNT_init (int* argc, char*** argv)
{
	int i, j, errcode;
	char mapName[128];
	char lpszNumProcs[128];
	char lpszFileMap[128];
	SECURITY_ATTRIBUTES saThread;
	DWORD SendThreadID;
	DWORD RecvThreadID;
	LONG nStatus;
	BOOL bStatus;
	/*DWORD dwStatus;*/

	HANDLE* hFileMaps;
	LPVOID* lpProcsBufsView;
	 
	/* Get rank from the  command line */
	MPID_MyWorldRank = atol((*argv)[*argc-1]);

	/* Get world size from the environment */
	errcode = GetEnvironmentVariable("MPID_WorldSize", lpszNumProcs, 127); 	
	MPID_WorldSize = atol(lpszNumProcs);

	_ASSERTE(MPID_WorldSize <= MPI_MAX_WORLD_SIZE);
	if (MPID_WorldSize > MPI_MAX_WORLD_SIZE) 
	{
		fprintf (stderr, "Too many processes in MPI_COMM_WORLD! Aborting the program!\n");
		MPID_WINNT_abort();
	}

	/* An array of filemapping objects */
	__MALLOC(hFileMaps, HANDLE, MPID_WorldSize);

	/* A global array of pointers to the shared memory space for every process */
	__MALLOC(lpProcsBufsView, LPVOID, MPID_WorldSize);

	/* Global 2D array of recv buffers for every process (1st index) to receive packets
	   from eevry other process (2nd index) */ 
	__MALLOC(WINNT_ShmemBufs, MPID_WINNT_SHMEM_BUF_T*, MPID_WorldSize);

	/* Note that we've only allocated memory for the pointers (2D arrays), we will allocate 
	   space for actual elements of the arrays below */

	for (i=0; i<MPID_WorldSize; i++) {
		/* Get file mapping handle */
		sprintf(mapName, "Proc_%d_bufferHandle", i);
		errcode = GetEnvironmentVariable(mapName, 
										 lpszFileMap,
										 127);
		hFileMaps[i] = (HANDLE) atol(lpszFileMap);

		/* Map file mapping into the process address space */
		lpProcsBufsView[i] = (LPVOID) MapViewOfFile(hFileMaps[i], 
													FILE_MAP_ALL_ACCESS,
													0,
													0,
													MPID_WorldSize*sizeof(MPID_WINNT_SHMEM_BUF_T));
   		if (lpProcsBufsView[i] == NULL) {
			fprintf(stderr, "Can't map process_%d buffers!", i);
			exit(1);
		}
		/* The memory for the shared buffers is allocated here in function MapViewOfFile */

		/* Tune base addresses of the arrays of shmem buffers appropriately */
		WINNT_ShmemBufs[i] = (MPID_WINNT_SHMEM_BUF_T*) lpProcsBufsView[i];

		_ASSERTE(_CrtIsValidPointer(WINNT_ShmemBufs[i],
									MPID_WorldSize*sizeof(MPID_WINNT_SHMEM_BUF_T),
									TRUE));
		/* Initialize buffer fields */
		for (j=0; j<MPID_WorldSize; j++) {	
			WINNT_ShmemBufs[i][j].Packet.head.lrank = -1;
			WINNT_ShmemBufs[i][j].Packet.head.tag = -1;
			WINNT_ShmemBufs[i][j].Packet.head.len = -1;
			/* Mutex and event handles are already there; we can access them */
		}

	}

	/* Open global event and semaphore for barrier syncronization */

	hBarrierEvent = OpenEvent(EVENT_ALL_ACCESS, 
								TRUE, 
								"__GLOBAL_BARRIER_EVENT__");
	_ASSERTE(hBarrierEvent);

	if(NULL == hBarrierEvent) {
		fprintf(stderr, "Fail to open a barrier event...\n");
		fprintf(stderr, "Aborting the program.\n");
		exit(1);	
	}
	
	hBarrierSemaphore = OpenSemaphore(SEMAPHORE_ALL_ACCESS, 
										TRUE,
										"__GLOBAL_BARRIER_SEMAPHORE__");
	_ASSERTE(hBarrierSemaphore);

	if(NULL == hBarrierSemaphore) {
		fprintf(stderr, "Fail to open a barrier semaphore...\n");
		fprintf(stderr, "Aborting the program.\n");
		exit(1);
	}

	/* Open global terminator event for abnormal termination */
	
	hTerminatorEvent = OpenEvent(EVENT_ALL_ACCESS, 
									TRUE, 
									"__GLOBAL_TERMINATOR_EVENT__");	
	_ASSERTE(hTerminatorEvent);

	if(NULL == hTerminatorEvent) {
		DWORD errcode = GetLastError();
		fprintf(stderr, "Fail to open a terminator event... error %d.\n", errcode);
		fprintf(stderr, "Aborting the program.\n");
		exit(1);	
	}

	/* Create an event to signal the end of send/recv operations */

	hEndProcessEvent = CreateEvent(NULL, TRUE, FALSE, NULL);

	_ASSERTE(hEndProcessEvent);

	if(NULL == hEndProcessEvent) {
		fprintf(stderr, "Fail to open an event...\n");
		fprintf(stderr, "Aborting the program.\n");
		exit(1);	
	}

	/* Initialize send and receive queues*/
	
	bStatus = PacketQueuesInitialize(&nStatus);
	
	_ASSERTE(bStatus);

	/* Fire of sender and receiver threads which will work between shared buffers and 
		queues of complete messages */

	saThread.nLength = sizeof(saThread);
	saThread.lpSecurityDescriptor = NULL;
	saThread.bInheritHandle = TRUE;

	hThreads[0] = CreateThread(&saThread, 
					0, /* default stack size */
					SHMEMReceiveThread,
					WINNT_ShmemBufs[MPID_MyWorldRank],
					0, /* started immediately */
					&SendThreadID);
	
	if(NULL == hThreads[0]) {
		fprintf(stderr, "Fail to create a receive thread...\n");
		fprintf(stderr, "Aborting the program.\n");
		exit(1);	
	}

	hThreads[1] = CreateThread(&saThread, 
					0, /* default stack size */
					SHMEMSendThread,
					WINNT_ShmemBufs,
					0, /* started immediately */
					&RecvThreadID);

	if(NULL == hThreads[1]) {
		fprintf(stderr, "Fail to create a send thread...\n");
		fprintf(stderr, "Aborting the program.\n");
		exit(1);	
	}

	hThreads[2] = CreateThread(&saThread, 
					0, /* default stack size */
					GLOBALTerminator,
					NULL,
					0, /* started immediately */
					&RecvThreadID);

	if(NULL == hThreads[2]) {
		fprintf(stderr, "Fail to create a send thread...\n");
		fprintf(stderr, "Aborting the program.\n");
		exit(1);	
	}

	/* Wait for all processes to pass their MPID_Init code */
	MPID_WINNT_Barrier();

	return 0;
}

void* MPID_WINNT_WorldSize(int* size)
{
	*size = MPID_WorldSize;
	return 0;
}

void* MPID_WINNT_MyRank(int* rank)
{
	*rank = MPID_MyWorldRank;
	return 0;
}

void MPID_WINNT_finalize()
{
	int i;
	DWORD dwStatus;
	BOOL bRetval;

	#if defined(SHMEM_ONLY) || defined(SHMEM_AND_TCP)
	while(!IsEmptyQueue(SHMEMSENDQUEUE))
		Sleep(50);
	#endif
	#if defined(TCP_ONLY) || defined(SHMEM_AND_TCP)
	while(!IsEmptyQueue(TCPIPSENDQUEUE))
		Sleep(50);
	#endif

	bRetval = SetEvent(hEndProcessEvent);
	
	_ASSERTE(bRetval);

	dwStatus = 	WaitForMultipleObjects(3, 
										hThreads,
										TRUE,
										INFINITE);

	_ASSERTE(WAIT_FAILED != dwStatus);

	for (i=0; i<MPID_WorldSize; i++)
		UnmapViewOfFile(WINNT_ShmemBufs[i]);

	bRetval = CloseHandle(hBarrierEvent);
	_ASSERTE(bRetval);
		bRetval = CloseHandle(hBarrierSemaphore);
	_ASSERTE(bRetval);
		bRetval = CloseHandle(hThreads[0]);
	_ASSERTE(bRetval);
		bRetval = CloseHandle(hThreads[1]);
	_ASSERTE(bRetval);
	bRetval = CloseHandle(hEndProcessEvent);
	_ASSERTE(bRetval);

	fflush(stderr);
	exit(0);
}

void MPID_WINNT_Barrier()
{
	DWORD dwStatus;
	BOOL bStatus;
/*
WAIT_OBJECT_0	The state of the specified object is signaled.
WAIT_TIMEOUT	The time-out interval elapsed, and the object's state is nonsignaled.

For a manual-reset event object, all waiting threads that can be released immediately are released. 
The function [PulseEvent] then resets the event object's state to nonsignaled and returns. 
	
*/
	dwStatus = WaitForSingleObject(hBarrierSemaphore, 0);

	_ASSERTE(WAIT_FAILED != dwStatus);

	if (WAIT_OBJECT_0 == dwStatus)
	{
		dwStatus = WaitForSingleObject(hBarrierEvent, INFINITE);
		_ASSERTE(WAIT_FAILED != dwStatus);
	}
	else 
	{
		bStatus = ReleaseSemaphore(hBarrierSemaphore, MPID_WorldSize - 1, NULL);
		_ASSERTE(bStatus); 
		bStatus = PulseEvent(hBarrierEvent);
		_ASSERTE(bStatus); 
	}
}

void MPID_WINNT_abort()
{
	BOOL bStatus;

	bStatus = SetEvent(hTerminatorEvent);
	_ASSERTE(bStatus);
	Sleep(INFINITE);
}

DWORD WINAPI GLOBALTerminator (LPVOID lpV)
{
	char msg[128];
	HANDLE hEvents[2]; 
	DWORD dwWhichEvent;
	hEvents[0] = hEndProcessEvent;
	hEvents[1] = hTerminatorEvent;


	dwWhichEvent = WaitForMultipleObjects(2,
										hEvents, 
										FALSE,
										INFINITE);
	_ASSERTE(WAIT_FAILED != dwWhichEvent);
		
	
	if (WAIT_OBJECT_0 == dwWhichEvent)
	{
		OutputDebugString("Normal exit of the GLOBALTerminator thread");
		ExitThread(0);
	}
	else
	{

		sprintf (msg, "Process [%d]: the GLOBALTerminator thread got termination signal.Aborting the process.", MPID_MyWorldRank);

		OutputDebugString(msg);
		ExitProcess(0xFFFFFFFF);
	}
	
	return 0xFFFFFFFF;
}	
